/*
    HD44780U and matrix keyboart functions
    Copyright (C) 2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <ftdi.h>


#include "error.h"
#include "hd.h"
#include "misc.h"
#include "zmem.h"


#define PIN(pins, value) if (value) hd->wr |= (pins); else hd->wr &= ~(pins);

#define WRITE \
    ret = ftdi_write_data(hd->ftdi, &hd->wr, 1); \
    if (ret){ \
        ftdierrormsg(hd->ftdi, "ftdi_write_data failed"); \
            return ret; \
    }

#define READ \
    ret = ftdi_read_data(hd->ftdi, &hd->rd, 1); \
    if (ret){ \
        ftdierrormsg(hd->ftdi, "ftdi_read_data failed");  \
        return ret; \
    }


struct hd *hd_init(int vid, int pid){
    struct hd *hd;
    int ret;

    hd = z_malloc(sizeof(struct hd));
    memset(hd, 0, sizeof(struct hd));

    ret = ftdi_usb_open(hd->ftdi, vid, pid);
    if (ret){ ftdierrormsg(hd->ftdi, "ftdi_usb_open failed");  return NULL; }

    ret = hd_reset(hd);
    if (ret) goto x;



    return hd; 
x:;
    ftdi_free(hd->ftdi);
    return NULL;  
}

int hd_reset(struct hd *hd){
    int ret;
    hd->wr &= ~(PIN_HD_D7|PIN_HD_D6|PIN_HD_D5|PIN_HD_D4|PIN_HD_E|PIN_HD_RW|PIN_HD_RS);
    hd->wr |= PIN_HD_BUSY;
    
    WRITE;
    READ;
    if (ret) return ret;
    usleep(16000); // more than 15 ms

    ret = hd_cmd_nowait(hd, 0x3);  // - - - -   0 0 1 DL 
    if (ret) return ret;
    usleep(5000);  // more than 4.1 ms

    ret = hd_cmd_nowait(hd, 0x3);
    if (ret) return ret;
    usleep(1000);  // more than 100 us

    ret = hd_cmd_nowait(hd, 0x3);
    if (ret) return ret;
    usleep(6000);  

    ret = hd_cmd_nowait(hd, 0x2);  // - - - -   0 0 1 DL    datalen
    if (ret) return ret;
    usleep(10000);

    ret = hd_cmd(hd, 0x28); // 0 0 1 DL  N F - -     datalen numberlines font
    if (ret) return ret;
    ret = hd_cmd(hd, 0x0c); // 0 0 0 0   1 D C B     displayon cursoron blinkcursor
    if (ret) return ret;
    ret = hd_cmd(hd, 0x01); // 0 0 0 0   0 0 0 1     clear display, DDRAM addr=0
    if (ret) return ret;
    ret = hd_cmd(hd, 0x06); // 0 0 0 0   0 1 I/D S   increment shift
    if (ret) return ret;

    return 0;
}


int hd_setdir(struct hd *hd){
    int ret;

    hd->bitmask |= (PIN_HD_D4 | PIN_HD_D5 | PIN_HD_D6 | PIN_HD_D7);
    ret = ftdi_set_bitmode(hd->ftdi, BITMODE_SYNCBB, hd->bitmask);
    return ret;
}

int hd_cmd_nowait(struct hd *hd, char c){
    int ret;
    
    PIN(PIN_HD_RS, 0);
    PIN(PIN_HD_RW, 0);
    PIN(PIN_HD_E, 0);
    WRITE;

    PIN(PIN_HD_D4, c & 0x01);
    PIN(PIN_HD_D5, c & 0x02);
    PIN(PIN_HD_D6, c & 0x04);
    PIN(PIN_HD_D7, c & 0x08);
    PIN(PIN_HD_E, 1);
    WRITE;

    PIN(PIN_HD_E, 0);
    WRITE;
    READ;
    READ;
    READ;

    return 0;
}

int hd_write(struct hd *hd, char c){
    int ret;

    PIN(PIN_HD_RW, 0);
    PIN(PIN_HD_E, 0);
    WRITE;

    PIN(PIN_HD_D4, c & 0x10);
    PIN(PIN_HD_D5, c & 0x20);
    PIN(PIN_HD_D6, c & 0x40);
    PIN(PIN_HD_D7, c & 0x80);
    PIN(PIN_HD_E, 1);
    WRITE;

    PIN(PIN_HD_E, 0);
    WRITE;

    PIN(PIN_HD_D4, c & 0x01);
    PIN(PIN_HD_D5, c & 0x02);
    PIN(PIN_HD_D6, c & 0x04);
    PIN(PIN_HD_D7, c & 0x08);
    PIN(PIN_HD_E, 1);
    WRITE;
    
    PIN(PIN_HD_E, 0);
    WRITE;

    READ;
    READ;
    READ;
    READ;
    READ;

    return 0;
}

int hd_wait(struct hd *hd){
    int ret, busy = 1;
    struct timeval start, stop;
    long int diff;

    gettimeofday(&start, NULL);
    while(busy){
        PIN(PIN_HD_RS, 0);
        PIN(PIN_HD_RW, 1); // read
        PIN(PIN_HD_E, 0);
        WRITE;
        
        PIN(PIN_HD_E, 1);
        WRITE;

        PIN(PIN_HD_E, 0);
        WRITE;
        
        PIN(PIN_HD_E, 1);
        WRITE;

        PIN(PIN_HD_E, 0);
        WRITE;

        READ;
        READ;
        READ;
        busy = hd->rd & PIN_HD_BUSY;
        READ;
        READ;

        gettimeofday(&stop, NULL);
        diff = difftimeval_ms(&stop, &start);
        if (diff > 5) {
            error("hd_wait: no BUSY reply\n");
            break;
        }
    }
    return 0;
}
                                

int hd_cmd(struct hd *hd, char c){
    int ret;

    PIN(PIN_HD_RS, 0);
    ret = hd_write(hd, c);
    if (ret) return ret;

    ret = hd_wait(hd);
    if (ret) return ret;
    return 0;
}

int hd_data(struct hd *hd, char c){
    int ret;

    PIN(PIN_HD_RS, 1);
    ret = hd_write(hd, c);
    if (ret) return ret;

    ret = hd_wait(hd);
    if (ret) return ret;

    return 0;
}

void hd_cgram(struct hd *hd, char pos) {
    hd_cmd(hd, 0x40|(pos&0x7f));
    return;
}

void hd_print(struct hd *hd, char *str){
    while (*str) {
        hd_data(hd, *str);
        str++;
    }
} 

void hd_print2(struct hd *hd, char pos, char *str){
    
    hd_cursor(hd, pos);
    while(*str){
        hd_data(hd, *str);
        str++;
        pos++;
#ifdef MC1601       
        if (pos==0+8) hd_cursor(hd, pos=64);
        if (pos==64+8) hd_cursor(hd, pos=0);
#endif      
    }
}

void hd_clear(struct hd *hd){
    unsigned char i;
    
    hd_cursor(hd, LINE_1);
    for (i=20;i;i--) {
        hd_data(hd, ' ');
    }
    
    hd_cursor(hd, LINE_2);
    for (i=20;i;i--) {
        hd_data(hd, ' ');
    }
}


/* keyboard routines */
int hd_keyb(struct hd *hd){
    int ret, cnt, scan;

    hd->bitmask &= ~(PIN_HD_D4 | PIN_HD_D5 | PIN_HD_D6 | PIN_HD_D7);
    ret = ftdi_set_bitmode(hd->ftdi, BITMODE_SYNCBB, hd->bitmask);

    PIN(PIN_HD_D4,1);
    PIN(PIN_HD_D5,1);
    PIN(PIN_HD_D6,1);
    PIN(PIN_HD_D7,1);
    PIN(PIN_KEYB_A0,0);
    PIN(PIN_KEYB_A1,0);
    PIN(PIN_KEYB_A2,0);
    PIN(PIN_HD_E,0);  // already 0
    WRITE;

    PIN(PIN_KEYB_A0,1);
    PIN(PIN_KEYB_A1,0);
    WRITE;

    PIN(PIN_KEYB_A0,0);
    PIN(PIN_KEYB_A1,1);
    WRITE;

    PIN(PIN_KEYB_A0,0);
    PIN(PIN_KEYB_A1,0);
    WRITE;
    WRITE;


    cnt = 0;
    scan = 0xff;
    READ;
    READ;
    if (!(hd->rd & PIN_HD_D4)) { scan = '1'; cnt++; }
    if (!(hd->rd & PIN_HD_D5)) { scan = '2'; cnt++; }
    if (!(hd->rd & PIN_HD_D6)) { scan = '3'; cnt++; }
    if (!(hd->rd & PIN_HD_D7)) { scan = '4'; cnt++; }

    READ;
    if (!(hd->rd & PIN_HD_D4)) { scan = '5'; cnt++; }
    if (!(hd->rd & PIN_HD_D5)) { scan = '6'; cnt++; }
    if (!(hd->rd & PIN_HD_D6)) { scan = '7'; cnt++; }
    if (!(hd->rd & PIN_HD_D7)) { scan = '8'; cnt++; }

    READ;
    if (!(hd->rd & PIN_HD_D4)) { scan = '9'; cnt++; }
    if (!(hd->rd & PIN_HD_D5)) { scan = 'A'; cnt++; }
    if (!(hd->rd & PIN_HD_D6)) { scan = 'B'; cnt++; }
    if (!(hd->rd & PIN_HD_D7)) { scan = 'C'; cnt++; }

    READ;
    if (!(hd->rd & PIN_HD_D4)) { scan = 'D'; cnt++; }
    if (!(hd->rd & PIN_HD_D5)) { scan = 'E'; cnt++; }
    if (!(hd->rd & PIN_HD_D6)) { scan = 'F'; cnt++; }
    if (!(hd->rd & PIN_HD_D7)) { scan = '0'; cnt++; }

    if (cnt != 1) return 0xff;
    return scan;
}



         
