/*
    serial - serial line functions
    Copyright (C) 2009 Ladislav Vaiz <ok1zia@nagano.cz>
    Hamlib code is (C) 2000-2005 by Stephane Fillod and Frank Singleton

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "config.h"

#include <ctype.h>
#include <stdlib.h>

#ifdef HAVE_TERMIOS_H
#include <termios.h>
#else
#ifdef HAVE_TERMIO_H
#include <termio.h>
#else
#ifdef HAVE_SGTTY_H
#include <sgtty.h>
#else
#error add AC_CHECK_HEADERS(sgtty.h termios.h termio.h) to configure.in 
#endif

#endif
#endif

#include "error.h"
#include "serial.h"
#include "zstring.h"

int serial_parse_string(char *str, int *baudrate, int *bits, char *parity, int *stopbits){
    struct zstring *zs;
    char *c, *c2;

    zs = zs_strdup(str);
    for (c = zs->str; *c != '\0'; c++){
        if (*c == ',') *c = ';';
    }

    c = zs_tokenize(zs, 1);
    if (c) *baudrate = atoi(c);
    else *baudrate = 9600;
    if (*baudrate <= 300) *baudrate = 9600;

    c = zs_tokenize(zs, 0);
    c2 = zs_tokenize(zs, 0);
    if (c && c2){
        if (atoi(c)>0){
            *bits = atoi(c);
            *parity = tolower(*c2);
        }else{
            *bits = atoi(c2);
            *parity = tolower(*c);
        }
    }else{
        *bits = 8;
        *parity = 'n';
    }

    c = zs_tokenize(zs, 0);
    if (c) *stopbits = atoi(c);
    else *stopbits = 1;
       
    return 0; 
}

#ifdef HAVE_LIBFTDI
int serial_ftdi_set_line(struct ftdi_context *ftdi, char *str){
    int baudrate, bits, stopbits;
    char parity;
    int ret;
    enum ftdi_bits_type fbits;
    enum ftdi_stopbits_type fstopbits;
    enum ftdi_parity_type fparity;

    serial_parse_string(str, &baudrate, &bits, &parity, &stopbits);

    if (bits == 7) fbits = BITS_7;
    else fbits = BITS_8;

    switch (stopbits){
        case 15: fstopbits = STOP_BIT_15; break;
        case 2:  fstopbits = STOP_BIT_2; break;
        default: fstopbits = STOP_BIT_1; break;
    }

    switch (tolower(parity)){
        case 'o': fparity = ODD;   break;
        case 'e': fparity = EVEN;  break;
        case 'm': fparity = MARK;  break;
        case 's': fparity = SPACE; break;
        default:  fparity = NONE;  break;
    }

    ret = ftdi_set_line_property(ftdi, fbits, fstopbits, fparity);
    return ret;
}
#endif

/* parts of code are taken from hamlib */
int serial_set_line(int fd, char *str){
    int baudrate, bits, stopbits;
    char parity;

    serial_parse_string(str, &baudrate, &bits, &parity, &stopbits);

  /* There's a lib replacement for termios under Mingw */
#if defined(HAVE_TERMIOS_H)
    struct termios options;
#elif defined(HAVE_TERMIO_H)
    struct termio options;
#elif defined(HAVE_SGTTY_H)
    struct sgttyb sg;
#else
#error "No term control supported!"
#endif

  
#if defined(HAVE_TERMIOS_H)
    tcgetattr(fd, &options);
#elif defined(HAVE_TERMIO_H)
    IOCTL (fd, TCGETA, &options);
#else	/* sgtty */
    IOCTL (fd, TIOCGETP, &sg);
#endif

#ifdef HAVE_CFMAKERAW
    cfmakeraw(&options); /* Set serial port to RAW mode by default. */
#endif

  /*
   * Set the baud rates to requested values
   */

    switch(baudrate){
        case 300:
            baudrate = B300;		/* yikes... */
            break;
        case 1200:
            baudrate = B1200;
            break;
        case 2400:
            baudrate = B2400;
            break;
        case 4800:
            baudrate = B4800;
            break;
        case 9600:
            baudrate = B9600;
            break;
        case 19200:
            baudrate = B19200;
            break;
        case 38400:
            baudrate = B38400;
            break;
        case 57600:
            baudrate = B57600;		/* cool.. */
            break;
        case 115200:
            baudrate = B115200;	/* awsome! */
            break;
        default:
            baudrate = B9600;
            break;
    }
  /* TODO */
    cfsetispeed(&options, baudrate);
    cfsetospeed(&options, baudrate);

/*
 * Enable the receiver and set local mode...
 */
  
    options.c_cflag |= (CLOCAL | CREAD);

/*
 * Set data to requested values.
 *
 */

    if (bits == 7){
        options.c_cflag &= ~CSIZE;
        options.c_cflag |= CS7;
    }else{
        options.c_cflag &= ~CSIZE;
        options.c_cflag |= CS8;
    }

/*
 * Set stop bits to requested values.
 *
 */  

    if (stopbits == 2){
        options.c_cflag |= CSTOPB;
    }else{
        options.c_cflag &= ~CSTOPB;
    }

/*
 * Set parity to requested values.
 *
 */  

    switch(parity){
        case 'e':      
            options.c_cflag |= PARENB;
            options.c_cflag &= ~PARODD;
            break;
        case 'o':
            options.c_cflag |= PARENB;
            options.c_cflag |= PARODD;
            break;
        default:
            options.c_cflag &= ~PARENB;
            break;
    }


/*
 * Set flow control to requested mode
 *
 */  
#if 0
  switch(rp->parm.serial.handshake) {
  case RIG_HANDSHAKE_NONE:
    options.c_cflag &= ~CRTSCTS;
    options.c_iflag &= ~IXON;
    break;
  case RIG_HANDSHAKE_XONXOFF:
    options.c_cflag &= ~CRTSCTS;
    options.c_iflag |= IXON;		/* Enable Xon/Xoff software handshaking */
    break;
  case RIG_HANDSHAKE_HARDWARE:
    options.c_cflag |= CRTSCTS;		/* Enable Hardware handshaking */
    options.c_iflag &= ~IXON;
    break;
  default:
    rig_debug(RIG_DEBUG_ERR, "open_serial: unsupported flow_control "
					"specified: %d\n",
					rp->parm.serial.handshake);
	CLOSE(fd);
    return -RIG_ECONF;
    break;
  }
#else
    options.c_cflag &= ~CRTSCTS;
    options.c_iflag &= ~IXON;
#endif

  /*
   * Choose raw input, no preprocessing please ..
   */

#if defined(HAVE_TERMIOS_H) || defined(HAVE_TERMIO_H)
    options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);

  /*
   * Choose raw output, no preprocessing please ..
   */

    options.c_oflag &= ~OPOST;

#else	/* sgtty */
    sg.sg_flags = RAW;
#endif


  /*
   * Flush serial port
   */

    tcflush(fd, TCIFLUSH);

  /*
   * Finally, set the new options for the port...
   */
  
#if defined(HAVE_TERMIOS_H)
    if (tcsetattr(fd, TCSANOW, &options) == -1) {
        errormsg("tcsetattr failed");
        return -1;
    }
#elif defined(HAVE_TERMIO_H)
    if (IOCTL(fd, TCSETA, &options) == -1) {
        errormsg("ioctl(TCSETA) failed");
        return -1;
    }
#else	/* sgtty */
    if (IOCTL(fd, TIOCSETP, &sg) == -1) {
        errormsg("ioctl(TIOCSETP) failed");
		return -1;
    }
#endif

    return 0;
}

#ifdef _WIN32
// http://msdn.microsoft.com/en-us/library/b0084kay(vs.71).aspx
int serial_win32_line(HANDLE handle, char *str){
    int baudrate, bits, stopbits;
    char parity;
    DCB dcb;

    serial_parse_string(str, &baudrate, &bits, &parity, &stopbits);

    dcb.fBinary=1;
    dcb.DCBlength=sizeof(DCB);

    dcb.BaudRate=9600;
    dcb.Parity=EVENPARITY;
    dcb.fParity=1;
    dcb.ByteSize=8;
    dcb.StopBits=0; //TODO 1;
#if 0    
    c='\0';
    if (strlen(s)>0) c=toupper(s[strlen(s)-1]);
    switch(c){
    case 'X':
        dcb->fInX=TRUE;
        dcb->fOutX=TRUE;
        dcb->fOutxDsrFlow=FALSE;
        dcb->fOutxCtsFlow=FALSE;
        dcb->fDtrControl=DTR_CONTROL_ENABLE;
        dcb->fRtsControl=RTS_CONTROL_ENABLE;
        break;
    case 'P':
        dcb->fInX=FALSE;
        dcb->fOutX=FALSE;
        dcb->fOutxDsrFlow=TRUE;
        dcb->fOutxCtsFlow=TRUE;
        dcb->fDtrControl=DTR_CONTROL_HANDSHAKE;
        dcb->fRtsControl=RTS_CONTROL_HANDSHAKE;
        break;
    default:
        dcb->fInX=FALSE;
        dcb->fOutX=FALSE;
        dcb->fOutxDsrFlow=FALSE;
        dcb->fOutxCtsFlow=FALSE;
        dcb->fDtrControl=DTR_CONTROL_ENABLE;
        dcb->fRtsControl=RTS_CONTROL_ENABLE;
        break;
    }
#endif

    dcb->BaudRate = baudrate;
    
    switch(tolower(parity)){
        case 'e':
            dcb->Parity=EVENPARITY;
            break;
        case 'm':
            dcb->Parity=MARKPARITY;
            break;
        case 'n':
            dcb->Parity=NOPARITY;
            break;
        case 'o':
            dcb->Parity=ODDPARITY;
            break;
        case 's':
            dcb->Parity=SPACEPARITY;
            break;
        default:
            dcb->Parity=NOPARITY;
            break;
    }

    dcb->ByteSize = bits;
    
    dcb->StopBits = stopbits;

    if (!SetCommState(handle, &dcb)) { 
        error("SetCommState failed err=%d\n", GetLastError());
        return -1;
    }

    return 0;
}

#endif
