
/*
    Copyright (C) 2009 Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

*/

#include "config.h"
#include "sound.h"
#include "error.h"
#include "zmem.h"

#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <unistd.h>

void sound_info(void){
#ifdef HAVE_ALSA
    alsa_info();
#endif   
}


const char *sound_frame_fmt2str(enum sound_frame_fmt fmt){
	switch(fmt){
		case SOUND_FMT_U8:
            return "U8";
        case SOUND_FMT_S16:
            return "S16";
        default:
            return "?";
    }
}

void sound_fmt_fill(struct sound_fmt *fmt){
	fmt->spf = fmt->channels;
	switch(fmt->fmt){
		case SOUND_FMT_U8:
			fmt->bps = 1;
			break;
		case SOUND_FMT_S16:
			fmt->bps = 2;
			break;
		default:
			fmt->bps = 1;
			break;
	}
    fmt->bpf = fmt->bps * fmt->spf;
    
	fmt->frame_time_us = 1000000L / fmt->rate;

	fmt->per_len_f = (fmt->per_time_us * fmt->rate) / 1000000L;
	fmt->per_len_s = fmt->per_len_f * fmt->spf;
	fmt->per_len_b = fmt->per_len_f * fmt->bpf;
}

snd_pcm_format_t sound_fmt2snd_pcm_format_t(struct sound_fmt *fmt){
	switch(fmt->fmt){
		case SOUND_FMT_U8:
            return SND_PCM_FORMAT_U8;
        case SOUND_FMT_S16:
            return SND_PCM_FORMAT_S16;
        default:
            return SND_PCM_FORMAT_S16;
    }
}


void sound_format_dump(struct sound_stream *stream, struct sound_fmt *fmt){
    if (fmt){
        dbg("f=%s ch=%d r=%d per_time_us=%d buf_time_us=%d\n", 
                sound_frame_fmt2str(fmt->fmt),
                fmt->channels, fmt->rate,
                fmt->per_time_us, fmt->buf_time_us);
        dbg("\tspf=%d bps=%d bpf=%d, frame_time=%dus\n", 
                fmt->spf, fmt->bps, fmt->bpf, 
                fmt->frame_time_us); 
        dbg("\tper_len=%df  =%ds  =%db\n", 
                fmt->per_len_f, fmt->per_len_s, fmt->per_len_b);
    }
    if (stream){
        dbg("\trate=%d  per_size=%d buf_size=%d  per_time_us=%d buf_time_us=%d\n", 
                stream->real_rate,
                stream->real_per_size, stream->real_buf_size,
                stream->real_per_time_us, stream->real_buf_time_us);
    }
}

struct sound *init_sound(enum sound_mode mode){
    struct sound *sound;

    sound = z_new0(struct sound);

    if (mode == SOUND_PLAY || mode == SOUND_DUPLEX){
        sound->play = init_sound_stream();
		sound->play->pcm_stream = SND_PCM_STREAM_PLAYBACK;
    }

    if (mode == SOUND_REC || mode == SOUND_DUPLEX){
        sound->rec = init_sound_stream();
		sound->rec->pcm_stream = SND_PCM_STREAM_CAPTURE;
	}

    return sound;
}

void free_sound(struct sound *sound){
    if (sound->play) {
        free_sound_stream(sound->play);
    }
    if (sound->rec) {
        free_sound_stream(sound->rec);
    }
    z_free(sound);
}

struct sound_stream *init_sound_stream(){
    struct sound_stream *stream;
    
    stream = z_new0(struct sound_stream);
    stream->sndfmt = z_new0(struct sound_fmt);    
    stream->datafmt = z_new0(struct sound_fmt);    
    return stream;
}

void free_sound_stream(struct sound_stream *stream){
    sound_stream_close(stream);
	z_free(stream->sndfmt);
	z_free(stream->datafmt);
	z_free(stream);
}

#define CHANNELS 3
int sound_stream_open(struct sound_stream *stream, const char *device){
	int ret, tmp, ratedir;
    unsigned int rateval;
    snd_pcm_hw_params_t *hwparams;
    snd_pcm_format_t format;
    int i;
    int channels[CHANNELS];
    unsigned int us;

    
   /* dbg("size=%d %d\n", sizeof(struct sound_stream), sizeof(struct sound_fmt));
    dbg("%p %d\n", stream, sizeof(struct sound_stream));
    dbg("%p %d %d\n", stream->sndfmt, sizeof(struct sound_fmt),
            (char*)stream->sndfmt - (char*)stream);
    dbg("%p %d %d\n", stream->datafmt, sizeof(struct sound_fmt),
            (char *)stream->datafmt - (char *)stream->sndfmt);*/
    sound_fmt_fill(stream->datafmt);
/*        dbg("\n");
        sound_format_dump(stream, stream->datafmt);
        dbg("\n");
        sound_format_dump(stream, stream->sndfmt);*/
    memcpy(stream->sndfmt, stream->datafmt, sizeof(struct sound_fmt));
/*        dbg("\n");
        sound_format_dump(stream, stream->datafmt);
        dbg("\n");
        sound_format_dump(stream, stream->sndfmt);
        exit(0);*/
	if (stream->opened){
        error("sound_stream_open: PCM already opened\n");
        return -1;
    }
        

	ret = snd_pcm_open(&stream->pcm, device, stream->pcm_stream, 0);
//    dbg("open: stream->pcm=%p\n", stream->pcm);
	if (ret < 0){
		snderrormsg(ret, "Can't open PCM '%s'", device);
		goto x;
	}

        
	stream->opened = 1;
	stream->recover = 5;
    
    snd_pcm_t *pcm = stream->pcm;
    snd_pcm_hw_params_alloca(&hwparams);
    
    ret = snd_pcm_hw_params_any(pcm, hwparams);
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params_any()");
        goto x;
    }  

    /* read/write format */
    ret = snd_pcm_hw_params_set_access(pcm, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED);
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params_set_access()");
        goto x;
    }  
    
    /* sample format */
    format = sound_fmt2snd_pcm_format_t(stream->datafmt);
    ret = snd_pcm_hw_params_set_format(pcm, hwparams, format);
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params_set_format(format=%d)", format);
        goto x;
    }  
    
    /* channels */
    channels[0] = stream->datafmt->channels;
    channels[1] = 2;
    channels[2] = 1;
 //       channels[0]=2;
    ret = 0;
    for (i=0; i<CHANNELS; i++){
        stream->sndfmt->channels = channels[i];
        ret = snd_pcm_hw_params_set_channels(pcm, hwparams, stream->sndfmt->channels);
        if (ret < 0) continue;
        dbg("sound_stream_open(%s): setting channels to %d\n", device, stream->sndfmt->channels);
            break;
    }
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params_set_channels(channels=%d)", stream->datafmt->channels);
        goto x;
    }  
    

    /* sample rate */
    rateval = stream->datafmt->rate;
    ratedir = 0;
    ret = snd_pcm_hw_params_set_rate_near(pcm, hwparams, &rateval, &ratedir);
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params_set_rate_near()");
        goto x;
    }  

    tmp = 1;
    ret = snd_pcm_hw_params_get_rate(hwparams, &stream->real_rate, &tmp); 
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params_get_rate()");
        goto x;
    }  
        

    /* buffer time */
    tmp = 1; /* not sure if tmp is input parameter. */
    us = stream->sndfmt->buf_time_us;
	ret = snd_pcm_hw_params_set_buffer_time_near(pcm, hwparams, &us, &tmp);
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params_set_buffer_time_near(%d)", stream->sndfmt->buf_time_us);
        goto x;
    }  

	ret = snd_pcm_hw_params_get_buffer_size(hwparams, &stream->real_buf_size);
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params_get_buffer_size()");
        goto x;
    }    
    
    tmp = 1;
	ret = snd_pcm_hw_params_get_buffer_time(hwparams, &stream->real_buf_time_us, &tmp);
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params_get_buffer_time()");
        goto x;
    }    
    
    /* period time */
    us = stream->sndfmt->per_time_us;
	ret = snd_pcm_hw_params_set_period_time_near(pcm, hwparams, &us, &tmp);
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params_set_period_time_near()");
        goto x;
    }  

	ret = snd_pcm_hw_params_get_period_size(hwparams, &stream->real_per_size, &tmp);
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params_get_period_size()");
        goto x;
    }  

    tmp = 1;
	ret = snd_pcm_hw_params_get_period_time(hwparams, &stream->real_per_time_us, &tmp);
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params_get_period_time()");
        goto x;
    }    
    
    

    /* write hwparams */
    ret = snd_pcm_hw_params(pcm, hwparams);
    if (ret < 0){
        snderrormsg(ret, "Failed snd_pcm_hw_params()");
        goto x;
    }  
    /*dbg("real values: rate=%d   buffer_time=%dus buffer_size=%d   period_time=%dus period_size=%d\n", 
            (int)rateval, 
            (int)stream->datafmt->buf_time_us, (int)stream->datafmt->buf_size, 
            (int)stream->datafmt->per_time_us, (int)stream->datafmt->per_size);
*/    
    sound_fmt_fill(stream->sndfmt);
   // sound_fmt_fill(stream->datafmt);
    return 0;
x:;    
    sound_stream_close(stream);
    return -1;
}

int sound_stream_close(struct sound_stream *stream){
    if (!stream->opened) return 0;
    dbg("close: stream->pcm=%p\n", stream->pcm);
    snd_pcm_close(stream->pcm);
    stream->opened = 0;
    return 0;
}


int sound_convert(void *srcbuf, void *dstbuf, struct sound_fmt *srcfmt, struct sound_fmt *dstfmt){
    if (srcfmt->channels == 1 && 
        dstfmt->channels == 2){
        int16_t *src, *dst;
        int i;

        src = (int16_t*)srcbuf;
        dst = (int16_t*)dstbuf;

        for (i=0; i < dstfmt->per_len_f; i++)
        {
            *dst++ = *src;   
            *dst++ = *src++;   
        }
    }

    if (srcfmt->channels == 2 && 
        dstfmt->channels == 1){
        int16_t *src, *dst;
        int i;

        src = (int16_t*)srcbuf;
        dst = (int16_t*)dstbuf;

        for (i=0; i < dstfmt->per_len_f; i++)
        {
            *dst = *src++/2;   
            *dst++ += *src++/2; 
        }
    }
    return 0;
}

int sound_playback(struct sound_stream *stream, void *srcbuf, int len_f){
    int ret;
    void *dstbuf = NULL, *buf;
    
    if (!stream->opened) return -1;
    buf = srcbuf;

    if (stream->datafmt->channels != stream->sndfmt->channels){
        dstbuf = z_malloc(stream->sndfmt->per_len_b);
        sound_convert(srcbuf, dstbuf, stream->datafmt, stream->sndfmt);
        buf = dstbuf;
    }

    ret = snd_pcm_writei(stream->pcm, buf, len_f);
//    dbg("snd_pcm_writei(%d frames)=%d\n", len_f, ret);
    if (ret < 0){
        if (ret == -EPIPE && stream->recover-- > 0){
            dbg("sound_playback: snd_pcm_writei returns %d %s, trying to recover... %d\n", ret, snd_strerror(ret), stream->recover);
            ret = snd_pcm_prepare(stream->pcm);    /* snd_pcm_recover is since libasound 1.0.11 */
        }
    }else{
        stream->recover = ALSA_RECOVER;
    }
    if (dstbuf) z_free(dstbuf);
    return ret;
}

int sound_record(struct sound_stream *stream, void *dstbuf, int len_f){
    int ret;
    void *srcbuf = NULL, *buf;

    if (!stream->opened) return -1;
    buf = dstbuf;
    
    if (stream->datafmt->channels != stream->sndfmt->channels){
        srcbuf = z_malloc(stream->sndfmt->per_len_b);
    }
        
    ret = snd_pcm_readi(stream->pcm, buf, len_f);
//    dbg("snd_pcm_readi(samples=%d) = %d\n", len_f, ret);
    if (ret < 0){
        if (ret == -EPIPE && stream->recover-- > 0){
            dbg("sound_record: snd_pcm_readi returns %d %s, trying to recover... %d\n", ret, snd_strerror(ret), stream->recover);
            ret = snd_pcm_prepare(stream->pcm);    /* snd_pcm_recover is since libasound 1.0.11 */
        }else{
            snderrormsg(ret, "sound_record: snd_pcm_readi failed");
        }
    }else{
        stream->recover = ALSA_RECOVER;
    }

    if (ret == -EPIPE){
        ret = 1;
        error("sound_record: alsa buffer overrun (%d), forcing return value %d\n", -EPIPE, ret); 
        usleep(10000);//fixme
    }
    
    if (srcbuf){
        sound_convert(srcbuf, dstbuf, stream->sndfmt, stream->datafmt);
        z_free(srcbuf);
    }

    return ret;
}

int sound_avail(struct sound_stream *stream){
    int ret = -1;

    snd_pcm_sframes_t frames;
    
    if (!stream->opened) return -1;
        
    frames = snd_pcm_avail_update(stream->pcm);
    if (frames < 0){
        if (frames == -EPIPE) return 1000000;
        snderrormsg(frames, "Failed snd_pcm_avail_update()");
        return frames;
    }  

/*       snd_pcm_status_t *status;
    snd_pcm_uframes_t uframes;
    
    snd_pcm_status_alloca(&status);
    snd_pcm_status(sound->alsa_play, status);
    uframes = snd_pcm_status_get_avail_max(status);*/
//        dbg("frames=%d samples\n", frames);
    ret = frames;
    return ret;
}

